"""Pycln CLI implementation."""
from pathlib import Path
from typing import Generator, List, Optional

import typer

from . import __doc__, __name__, version_callback
from .utils import pathu, refactor, regexu, report
from .utils.config import Config

app = typer.Typer(name=__name__, add_completion=True)


@app.command(context_settings=dict(help_option_names=["-h", "--help"]))
def main(  # pylint: disable=R0913,R0914
    paths: List[Path] = typer.Argument(None, help="Directories or files paths."),
    config: Optional[Path] = typer.Option(
        None,
        "--config",
        show_default=False,
        help="Read configuration from a file.",
    ),
    include: str = typer.Option(
        regexu.INCLUDE_REGEX,
        "--include",
        "-i",
        show_default=True,
        help=(
            "A regular expression that matches files and directories"
            " that should be included on recursive searches."
            " An empty value means all files are included regardless of the name."
            " Use forward slashes for directories on all platforms (Windows, too)."
            " Exclusions are calculated first, inclusions later."
        ),
    ),
    exclude: str = typer.Option(
        regexu.EXCLUDE_REGEX,
        "--exclude",
        "-e",
        show_default=True,
        help=(
            "A regular expression that matches files and directories"
            " that should be exclude on recursive searches."
            " An empty value means no paths are excluded."
            " Use forward slashes for directories on all platforms (Windows, too)."
            " Exclusions are calculated first, inclusions later."
        ),
    ),
    all_: bool = typer.Option(
        False,
        "--all",
        "-a",
        show_default=True,
        help="Remove all unused imports (not just those checked from side effects).",
    ),
    check: bool = typer.Option(
        False,
        "--check",
        "-c",
        show_default=True,
        help=(
            "Do not write the files back, just return the status."
            " Return code 0 means nothing would change."
            " Return code 1 means some files would be changed."
            " Return code 250 means there was an internal error."
        ),
    ),
    diff: bool = typer.Option(
        False,
        "--diff",
        "-d",
        show_default=True,
        help="Do not write the files back, just output a diff for each file on stdout.",
    ),
    verbose: bool = typer.Option(
        False,
        "--verbose",
        "-v",
        show_default=True,
        help=(
            "Also emit messages to stderr about files"
            " that were not changed and about files/imports that were ignored."
        ),
    ),
    quiet: bool = typer.Option(
        False,
        "--quiet",
        "-q",
        show_default=True,
        help=(
            "Do not emit both removed and expanded imports "
            "and non-error messages to stderr."
            " Errors are still emitted;"
            " silence those with `-s, --silence`"
        ),
    ),
    silence: bool = typer.Option(
        False,
        "--silence",
        "-s",
        show_default=True,
        help=(
            "Silence both stdout and stderr."
            " Uncaught errors are sill emitted;"
            " silence those with 2>/dev/null."
            " (not recommended)."
        ),
    ),
    expand_stars: bool = typer.Option(
        False,
        "--expand-stars",
        "-x",
        help=(
            "Expand wildcard star imports."
            " It works if only if the module is importable."
        ),
    ),
    no_gitignore: bool = typer.Option(
        False,
        "--no-gitignore",
        show_default=True,
        help="Do not ignore `.gitignore` patterns. if present.",
    ),
    version: bool = typer.Option(  # pylint: disable=W0613
        None,
        "--version",
        callback=version_callback,
        help="Show the version and exit.",
    ),
):
    configs = Config(
        paths=paths,
        config=config,
        include=include,  # type: ignore
        exclude=exclude,  # type: ignore
        all_=all_,
        check=check,
        diff=diff,
        verbose=verbose,
        quiet=quiet,
        silence=silence,
        expand_stars=expand_stars,
        no_gitignore=no_gitignore,
    )
    reporter = report.Report(configs)
    session_maker = refactor.Refactor(configs, reporter)
    gitignore = regexu.get_gitignore(Path("."), configs.no_gitignore)
    for path in configs.paths:
        sources: Generator = pathu.yield_sources(
            path, configs.include, configs.exclude, gitignore, reporter
        )
        for source in sources:
            session_maker.session(source)
    # Print the report.
    typer.echo(str(reporter), nl=False)
    # Set the correct exit code and exit.
    typer.Exit(reporter.exit_code)


# Override main function `__doc__`.
# This `__doc__` has read from `pyproject.toml`.
main.__doc__ = __doc__



"""Pycln CST transforming utility."""
from pathlib import Path
from typing import List, Optional, Set, TypeVar, Union, cast

import libcst as cst

from ._exceptions import UnsupportedCase
from ._nodes import NodeLocation

# Constants.
SPACE4 = " " * 4

# Custom types.
ImportT = TypeVar("ImportT", bound=Union[cst.Import, cst.ImportFrom])


class ImportTransformer(cst.CSTTransformer):

    """Import statements transformer.

    :param used_names: set of all used names to keep.
    :param location: `NodeLocation`.
    """

    def __init__(self, used_names: Set[str], location: NodeLocation):
        if not used_names:
            # Bad class usage.
            raise ValueError("'used_names' parameter can't be empty set.")
        self._used_names = used_names
        self._location = location
        self._indentation = " " * location.start.col

    def refactor_import_star(self, updated_node: cst.ImportFrom) -> cst.ImportFrom:
        """Add used import aliases to import star.

        :param updated_node: `cst.ImportFrom` node to refactor.
        :returns: refactored node.
        """
        is_multiline = len(self._used_names) > 3
        used_aliases: List[cst.ImportAlias] = []
        for name in self._used_names:

            # Skip any dotted name in order
            # to avoid names collision.
            if "." in name:
                continue

            # Initialy create a single line alias.
            cst_alias = cst.ImportAlias(
                name=cst.Name(name),
                comma=cst.Comma(whitespace_after=cst.SimpleWhitespace(" ")),
            )

            # Convert the single line alias to multiline
            # if there're more than 3 used names.
            if is_multiline:
                cst_alias = self._multiline_alias(cst_alias)

            used_aliases.append(cst_alias)

        return self._stylize(updated_node, used_aliases, is_multiline)

    def refactor_import(self, updated_node: ImportT) -> ImportT:
        """Remove unused imports from the given `updated_node`.

        :param updated_node: `cst.Import` or `cst.ImportFrom` node to refactor.
        :returns: refactored node.
        """
        used_aliases: List[cst.ImportAlias] = []
        for alias in updated_node.names:  # type: ignore
            if self._get_alias_name(alias.name) in self._used_names:
                used_aliases.append(alias)
        return self._stylize(updated_node, used_aliases)

    def leave_Import(  # pylint: disable=W0613
        self, original_node: cst.Import, updated_node: cst.Import
    ) -> Union[cst.Import]:
        return self.refactor_import(updated_node)

    def leave_ImportFrom(  # pylint: disable=W0613
        self, original_node: cst.ImportFrom, updated_node: cst.ImportFrom
    ) -> Union[cst.ImportFrom]:
        if isinstance(updated_node.names, cst.ImportStar):
            return self.refactor_import_star(updated_node)
        else:
            return self.refactor_import(updated_node)

    def _get_alias_name(
        self, node: Optional[Union[cst.Name, cst.Attribute]], name=""
    ) -> str:
        # Recursion function that calculates `node` string dotted name.
        if isinstance(node, cst.Name):
            name += node.value
            return name
        return self._get_alias_name(node.value) + "." + node.attr.value  # type: ignore

    @staticmethod
    def _multiline_parenthesized_whitespace(indent: str) -> cst.ParenthesizedWhitespace:
        # Return multiline parenthesized white space.
        return cst.ParenthesizedWhitespace(
            indent=True,
            last_line=cst.SimpleWhitespace(value=indent),
        )

    def _multiline_alias(self, alias: cst.ImportAlias) -> cst.ImportAlias:
        # Convert the given `alias` to multiline `alias`.
        return cst.ImportAlias(
            name=alias.name,
            asname=alias.asname,
            comma=cst.Comma(
                whitespace_after=ImportTransformer._multiline_parenthesized_whitespace(
                    self._indentation + SPACE4
                )
            ),
        )

    def _multiline_lpar(self) -> cst.LeftParen:
        # Return multiline `cst.LeftParen`.
        return cst.LeftParen(
            whitespace_after=ImportTransformer._multiline_parenthesized_whitespace(
                self._indentation + SPACE4
            )
        )

    def _multiline_rpar(self) -> cst.RightParen:
        # Return multiline `cst.RightParen`.
        return cst.RightParen(
            whitespace_before=ImportTransformer._multiline_parenthesized_whitespace(
                self._indentation
            )
        )

    def _stylize(
        self,
        node: ImportT,
        used_aliases: List[cst.ImportAlias],
        force_multiline: bool = False,
    ) -> ImportT:
        # (Preserving `node` style).
        # Remove the comma from the last name.
        used_aliases[-1] = used_aliases[-1].with_changes(
            comma=cst.MaybeSentinel.DEFAULT
        )
        node = cast(ImportT, node.with_changes(names=used_aliases))
        # Preserving multiline nodes style.
        if isinstance(node, cst.ImportFrom):
            if force_multiline or (node.rpar and len(self._location) != 1):
                rpar, lpar = self._multiline_rpar(), self._multiline_lpar()
                node = cast(ImportT, node.with_changes(rpar=rpar, lpar=lpar))
        return node


def rebuild_import(
    import_stmnt: str,
    used_names: Set[str],
    path: Path,
    location: NodeLocation,
) -> List[str]:
    """Rebuild the given `import_stmnt` based on `used_names` using `LibCST`.

    :param import_stmnt: source code of the import statement.
    :param used_names: set of all used names to keep.
    :param path: where `import_stats` has imported.
    :param location: `NodeLocation`.
    :returns: fixed import statement source code as list of lines.
    :raises cst.ParserSyntaxError: in some rare cases.
    :raises UnsupportedCase: in some rare cases.
    """
    if ";" in import_stmnt:
        msg = "import statements separated with ';'."
        raise UnsupportedCase(path, location, msg)

    # Remove `import_stmnt` indentation/last-"\n".
    stripped_stmnt = import_stmnt.lstrip(" ").rstrip("\n")
    indentation = " " * location.start.col

    # Remove unused aliases.
    fixed_lines: List[str] = []
    if used_names:
        transformer = ImportTransformer(used_names, location)
        cst_tree = cst.parse_module(stripped_stmnt)  # May raise cst.ParserSyntaxError.
        fixed_lines = cst_tree.visit(transformer).code.splitlines(keepends=True)

    if not fixed_lines:
        # Replace the removed import with a pass statement.
        fixed_lines = [f"{indentation}pass\n" if indentation else ""]
    else:
        # Reinsert the removed indentation.
        fixed_lines[0] = indentation + fixed_lines[0]

        # Reinsert the removed `"\n"`.
        fixed_lines[-1] = fixed_lines[-1] + "\n"

    return fixed_lines



"""Pycln source code AST analysis utility."""
import ast
import os
import sys
from dataclasses import dataclass
from enum import Enum, unique
from functools import wraps
from pathlib import Path
from typing import Any, Callable, List, Optional, Set, Tuple, TypeVar, Union, cast

from . import _nodes, iou, pathu
from ._exceptions import ReadPermissionError, UnexpandableImportStar, UnparsableFile

# Constants.
PY38_PLUS = sys.version_info >= (3, 8)
IMPORT_EXCEPTIONS = {"ImportError", "ImportWarning", "ModuleNotFoundError"}
__ALL__ = "__all__"
NAMES_TO_SKIP = {
    "__name__",
    "__doc__",
    "__package__",
    "__loader__",
    "__spec__",
    "__build_class__",
    "__import__",
    __ALL__,
}

# Custom types.
FunctionT = TypeVar("FunctionT", bound=Callable[..., Any])
FunctionDefT = TypeVar(
    "FunctionDefT", bound=Union[ast.FunctionDef, ast.AsyncFunctionDef]
)


def recursive(func: FunctionT) -> FunctionT:
    """decorator to make `ast.NodeVisitor` work recursive.

    :param func: `ast.NodeVisitor.visit_*` function.
    """

    @wraps(func)
    def wrapper(self, *args, **kwargs):
        func(self, *args, **kwargs)
        self.generic_visit(*args)

    return cast(FunctionT, wrapper)


@dataclass
class ImportStats:

    """Import statements statistics."""

    import_: Set[_nodes.Import]
    from_: Set[_nodes.ImportFrom]

    def __iter__(self):
        return iter([self.import_, self.from_])


@dataclass
class SourceStats:

    """Source code (`ast.Name`, `ast.Attribute`) statistics."""

    #: Included on `__iter__`.
    name_: Set[str]
    attr_: Set[str]

    #: Not included on `__iter__`.
    names_to_skip: Set[str]

    def __iter__(self):
        return iter([self.name_, self.attr_])


class SourceAnalyzer(ast.NodeVisitor):

    """AST source code analyzer.

    >>> import ast
    >>> source = "source.py"
    >>> with open(source, "r") as sourcef:
    >>>     source_lines = sourcef.readlines()
    >>>     tree = ast.parse("".join(source_lines))
    >>> analyzer = SourceAnalyzer(source_lines)
    >>> analyzer.visit(tree)
    >>> source_stats, import_stats = analyzer.get_stats()

    :param source_lines: source code as string lines,
        required only when Python < 3.8.
    :raises ValueError: when Python < 3.8 and no source code lines provided.
    """

    def __init__(self, source_lines: Optional[List[str]] = None):
        if not PY38_PLUS and source_lines is None:
            # Bad class usage.
            raise ValueError("Please provide source lines for Python < 3.8.")
        self._lines = source_lines
        self._import_stats = ImportStats(set(), set())
        self._imports_to_skip: Set[Union[_nodes.Import, _nodes.ImportFrom]] = set()
        self._source_stats = SourceStats(set(), set(), set())

    @recursive
    def visit_Import(self, node: ast.Import):
        if node not in self._imports_to_skip:
            py38_node = self._get_py38_import_node(node)
            self._import_stats.import_.add(py38_node)

    @recursive
    def visit_ImportFrom(self, node: ast.ImportFrom):
        if node not in self._imports_to_skip:
            py38_node = self._get_py38_import_from_node(node)
            if not str(py38_node.module).startswith("__"):
                self._import_stats.from_.add(py38_node)

    @recursive
    def visit_Name(self, node: ast.Name):
        self._source_stats.name_.add(node.id)

    @recursive
    def visit_Attribute(self, node: ast.Attribute):
        self._source_stats.attr_.add(node.attr)

    @recursive
    def visit_Try(self, node: ast.Try):
        """Support any try/except block that has import error(s).

        Add any import that placed on try/except block that has import error
        to `self._imports_to_skip`.

        exp:- two of these imports would not be used and should not be removed:
            >>> try:
            >>>     import foo2
            >>> except ModuleNotFoundError:
            >>>     import foo3
            >>> else:
            >>>     import foo38

        supported exceptions (`IMPORT_EXCEPTIONS`):
            - ModuleNotFoundError
            - ImportError.
            - ImportWarning.

        supported blocks:
            - try.
            - except.
            - else.
        """
        is_skip_case = False

        def add_imports_to_skip(body: List[ast.stmt]) -> None:
            """Add all try/except/else blocks body import children to
            `self._imports_to_skip`.

            :param body: ast.List to iterate over.
            """
            for child in body:
                if hasattr(child, "names"):
                    self._imports_to_skip.add(child)  # type: ignore

        for handler in node.handlers:
            if hasattr(handler.type, "elts"):
                for name in getattr(handler.type, "elts", []):
                    if hasattr(name, "id") and name.id in IMPORT_EXCEPTIONS:
                        is_skip_case = True
                        break
            elif hasattr(handler.type, "id"):
                if getattr(handler.type, "id", "") in IMPORT_EXCEPTIONS:
                    is_skip_case = True
            if is_skip_case:
                add_imports_to_skip(handler.body)

        if is_skip_case:
            for body in (node.body, node.orelse):
                add_imports_to_skip(body)

    @recursive
    def visit_AnnAssign(self, node: ast.AnnAssign):
        #: Support string type annotations.
        #: >>> from typing import List
        #: >>> foo: "List[str]" = []
        self._visit_string_type_annotation(node)

    @recursive
    def visit_arg(self, node: ast.arg):
        # Support Python ^3.8 type comments.
        self._visit_type_comment(node)
        #: Support arg string type annotations.
        #: >>> from typing import Tuple
        #: >>> def foo(bar: "Tuple[str, int]"):
        #: ...     pass
        self._visit_string_type_annotation(node)

    @recursive
    def visit_FunctionDef(self, node: FunctionDefT):
        # Support Python ^3.8 type comments.
        self._visit_type_comment(node)
        #: Support string type annotations.
        #: >>> from typing import List
        #: >>> def foo() -> 'List[str]':
        #: >>>     pass
        self._visit_string_type_annotation(node)

    # Support `ast.AsyncFunctionDef`.
    visit_AsyncFunctionDef = visit_FunctionDef

    @recursive
    def visit_Assign(self, node: ast.Assign):
        # Support Python ^3.8 type comments.
        self._visit_type_comment(node)
        id_ = getattr(node.targets[0], "id", None)
        # These names will be skipped on import `*` case.
        if id_ in NAMES_TO_SKIP:
            self._source_stats.names_to_skip.add(id_)
        # Support `__all__` dunder overriding case.
        if id_ == __ALL__:
            if isinstance(node.value, (ast.List, ast.Tuple, ast.Set)):
                for constant in node.value.elts:
                    k = "s" if hasattr(constant, "s") else "value"
                    value = getattr(constant, k, "")
                    if value and isinstance(value, str):
                        self._source_stats.name_.add(value)

    def _visit_string_type_annotation(
        self, node: Union[ast.AnnAssign, ast.arg, FunctionDefT]
    ) -> None:
        # Support string type annotations.
        if isinstance(node, (ast.AnnAssign, ast.arg)):
            annotation = node.annotation
        else:
            annotation = node.returns
        if isinstance(annotation, (ast.Constant, ast.Str)):
            if hasattr(annotation, "value"):
                value = annotation.value
            else:
                value = annotation.s
            if value:
                tree = parse_ast(value, mode="eval")
                self._add_name_attr(tree)

    def _visit_type_comment(
        self, node: Union[ast.Assign, ast.arg, FunctionDefT]
    ) -> None:
        #: Support Python ^3.8 type comments.
        #:
        #: This feature is only available for Python ^3.8.
        #: PEP 526 -- Syntax for Variable Annotations.
        #: For more information:
        #:     - https://www.python.org/dev/peps/pep-0526/
        #:     - https://docs.python.org/3.8/library/ast.html#ast.parse
        type_comment = getattr(node, "type_comment", None)
        if type_comment:
            if isinstance(node, (ast.Assign, ast.arg)):
                mode = "eval"
            else:
                mode = "func_type"
            tree = parse_ast(type_comment, mode=mode)
            self._add_name_attr(tree)

    def _add_name_attr(self, tree: ast.AST):
        # Add any `ast.Name` or `ast.Attribute`
        # child to `self._source_stats`.
        for node in ast.walk(tree):
            if isinstance(node, ast.Name):
                self._source_stats.name_.add(node.id)
            elif isinstance(node, ast.Attribute):
                self._source_stats.attr_.add(node.attr)

    def _get_py38_import_node(self, node: ast.Import) -> _nodes.Import:
        # Convert any Python < 3.8 `ast.Import`
        # to `_nodes.Import` in order to support `end_lineno`.
        if hasattr(node, "end_lineno"):
            end_lineno = node.end_lineno
        else:
            line = self._lines[node.lineno - 1]
            multiline = SourceAnalyzer._is_parentheses(line) is not None
            end_lineno = node.lineno + (1 if multiline else 0)
        location = _nodes.NodeLocation((node.lineno, node.col_offset), end_lineno)
        return _nodes.Import(location=location, names=node.names)

    def _get_py38_import_from_node(self, node: ast.ImportFrom) -> _nodes.ImportFrom:
        # Convert any Python < 3.8 `ast.ImportFrom`
        # to `_nodes.ImportFrom` in order to support `end_lineno`.
        if hasattr(node, "end_lineno"):
            end_lineno = node.end_lineno
        else:
            line = self._lines[node.lineno - 1]
            is_parentheses = SourceAnalyzer._is_parentheses(line)
            multiline = is_parentheses is not None
            end_lineno = (
                node.lineno
                if not multiline
                else self._get_end_lineno(node.lineno, is_parentheses)
            )
        location = _nodes.NodeLocation((node.lineno, node.col_offset), end_lineno)
        return _nodes.ImportFrom(
            location=location,
            names=node.names,
            module=node.module,
            level=node.level,
        )

    @staticmethod
    def _is_parentheses(import_from_line: str) -> Optional[bool]:
        # Return importFrom multi-line type.
        # ('(' => True), ('\\' => False) else None.
        if "(" in import_from_line:
            return True
        elif "\\" in import_from_line:
            return False
        else:
            return None

    def _get_end_lineno(self, lineno: int, is_parentheses: bool) -> int:
        # Get `ast.ImportFrom` `end_lineno` of the given `lineno`.
        lines_len = len(self._lines)
        for end_lineno in range(lineno, lines_len):
            if is_parentheses:
                if ")" in self._lines[end_lineno]:
                    end_lineno += 1
                    break
            else:
                if "\\" not in self._lines[end_lineno]:
                    end_lineno += 1
                    break
        return end_lineno

    def get_stats(self) -> Tuple[SourceStats, ImportStats]:
        """Get source analyzer results.

        :returns: tuple of `ImportStats` and `SourceStats`.
        """
        return self._source_stats, self._import_stats


class ImportablesAnalyzer(ast.NodeVisitor):

    """Get set of all importable names from given `ast.Module`.

    >>> import ast
    >>> source = "source.py"
    >>> with open(source, "r") as sourcef:
    >>>     tree = ast.parse(sourcef.read())
    >>> analyzer = ImportablesAnalyzer(source)
    >>> analyzer.visit(tree)
    >>> importable_names = analyzer.get_stats()

    :param path: a file path that belongs to the given `ast.Module`.
    """

    def __init__(self, path: Path):
        self._not_importables: Set[Union[ast.Name, str]] = set()
        self._importables: Set[str] = set()
        self._has_all = False
        self._path = path

    @recursive
    def visit_Assign(self, node: ast.Assign):
        # Support `__all__` dunder overriding case.
        id_ = getattr(node.targets[0], "id", None)
        if id_ == __ALL__:
            if isinstance(node.value, (ast.List, ast.Tuple, ast.Set)):
                self._has_all = True
                self._importables.clear()
                for constant in node.value.elts:
                    k = "s" if hasattr(constant, "s") else "value"
                    value = getattr(constant, k, "")
                    if value and isinstance(value, str):
                        self._importables.add(value)

    @recursive
    def visit_Import(self, node: ast.Import):
        # Analyze each import statement.
        for alias in node.names:
            name = alias.asname if alias.asname else alias.name
            self._importables.add(name)

    @recursive
    def visit_ImportFrom(self, node: ast.ImportFrom):
        # Analyze each importFrom statement.
        try:
            if node.names[0].name == "*":
                # Expand import star if possible.
                node = cast(ast.ImportFrom, expand_import_star(node, self._path))
            for alias in node.names:
                name = alias.asname if alias.asname else alias.name
                self._importables.add(name)
        except UnexpandableImportStar:  # pragma: no cover
            # * We shouldn't do anything because it's not importable.
            pass  # pragma: no cover

    @recursive
    def visit_FunctionDef(self, node: FunctionDefT):
        # Add function name as importable name.
        if node.name not in self._not_importables:
            self._importables.add(node.name)
        self._compute_not_importables(node)

    # Support `ast.AsyncFunctionDef`.
    visit_AsyncFunctionDef = visit_FunctionDef

    @recursive
    def visit_ClassDef(self, node: ast.ClassDef):
        # Add class name as importable name.
        if node.name not in self._not_importables:
            self._importables.add(node.name)
        self._compute_not_importables(node)

    @recursive
    def visit_Name(self, node: ast.Name):
        if isinstance(node.ctx, ast.Store):
            # Except not-importables.
            if node not in self._not_importables:
                self._importables.add(node.id)

    def _compute_not_importables(self, node: Union[FunctionDefT, ast.ClassDef]):
        # Compute class/function not-importables.
        for node_ in ast.iter_child_nodes(node):

            if isinstance(node_, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)):
                self._not_importables.add(cast(str, node_.name))

            if isinstance(node_, ast.Assign):
                for target in node_.targets:
                    self._not_importables.add(cast(ast.Name, target))

    def get_stats(self) -> Set[str]:
        if self._path.name == "__init__.py":
            for path in os.listdir(self._path.parent):
                file_path = self._path.parent.joinpath(path)
                if file_path.is_dir() or path.endswith(".py"):
                    self._importables.add(path.split(".")[0])
        return self._importables

    def generic_visit(self, node):
        """Called if no explicit visitor function exists for a node
        (override)."""
        # Continue visiting if only if `__all__` has not overridden.
        if not self._has_all:
            for _, value in ast.iter_fields(node):
                if isinstance(value, list):
                    for item in value:
                        if isinstance(item, ast.AST):
                            self.visit(item)
                elif isinstance(value, ast.AST):
                    self.visit(value)


@unique
class HasSideEffects(Enum):
    """SideEffects values."""

    YES = 1
    MAYBE = 0.5
    NO = 0

    #: Some names aren't modules.
    NOT_MODULE = -1

    #: Just in case an exception has raised
    #: while parsing a file.
    NOT_KNOWN = -2


class SideEffectsAnalyzer(ast.NodeVisitor):

    """Check if the given `ast.Module` has side effects or not.

    >>> import ast
    >>> source = "source.py"
    >>> with open(source, "r") as sourcef:
    >>>     tree = ast.parse(sourcef.read())
    >>> analyzer = SideEffectsAnalyzer()
    >>> analyzer.visit(tree)
    >>> has_side_effects = analyzer.has_side_effects()
    """

    def __init__(self):
        self._not_side_effects: Set[ast.Call] = set()
        self._has_side_effects = HasSideEffects.NO

    @recursive
    def visit_FunctionDef(self, node: FunctionDefT):
        # Mark any call inside a function as not-side-effect.
        self._compute_not_side_effects(node)

    # Support `ast.AsyncFunctionDef`.
    visit_AsyncFunctionDef = visit_FunctionDef

    @recursive
    def visit_ClassDef(self, node: ast.ClassDef):
        # Mark any call inside a class as not-side-effect.
        self._compute_not_side_effects(node)

    def _compute_not_side_effects(
        self, node: Union[ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef]
    ) -> None:
        # Mark any call inside the given `node` as not-side-effect.
        for node_ in ast.iter_child_nodes(node):
            if isinstance(node_, ast.Expr):
                if isinstance(node_.value, ast.Call):
                    self._not_side_effects.add(node_.value)

    @recursive
    def visit_Call(self, node: ast.Call):
        if node not in self._not_side_effects:
            self._has_side_effects = HasSideEffects.YES

    @recursive
    def visit_Import(self, node: ast.Import):
        self._has_side_effects = SideEffectsAnalyzer._check_names(node.names)

    @recursive
    def visit_ImportFrom(self, node: ast.ImportFrom):
        packages = node.module.split(".") if node.module else []
        packages_aliases = [ast.alias(name=name, asname=None) for name in packages]
        self._has_side_effects = SideEffectsAnalyzer._check_names(packages_aliases)
        if self._has_side_effects is HasSideEffects.NO:
            self._has_side_effects = SideEffectsAnalyzer._check_names(node.names)

    @staticmethod
    def _check_names(names: List[ast.alias]) -> HasSideEffects:
        # Check if imported names has side effects or not.
        for alias in names:

            # All standard lib modules doesn't has side effects
            # except `pathu.IMPORTS_WITH_SIDE_EFFECTS`.
            if alias.name in pathu.get_standard_lib_names():
                continue

            # Known side effects.
            if alias.name in pathu.IMPORTS_WITH_SIDE_EFFECTS:
                return HasSideEffects.YES

            # [Here instead of doing that, we can make the analyzer
            # works recursively inside each not known import.
            return HasSideEffects.MAYBE
            # I choosed this way because it's almost %100 we will end
            # with a file that has side effects :-) ].

        return HasSideEffects.NO

    def has_side_effects(self) -> HasSideEffects:
        return self._has_side_effects

    def generic_visit(self, node):
        """Called if no explicit visitor function exists for a node
        (override)."""
        # Continue visiting if only if there's no know side effects.
        if self._has_side_effects is HasSideEffects.NO:
            for _, value in ast.iter_fields(node):
                if isinstance(value, list):
                    for item in value:
                        if isinstance(item, ast.AST):
                            self.visit(item)
                elif isinstance(value, ast.AST):
                    self.visit(value)


def expand_import_star(
    node: Union[ast.ImportFrom, _nodes.ImportFrom], path: Path
) -> Union[ast.ImportFrom, _nodes.ImportFrom]:
    """Expand import star statement, replace the `*` with a list of ast.alias.

    :param node: `_nodes/ast.ImportFrom` node that has a '*' as `alias.name`.
    :param path: where the node has imported.
    :returns: expanded `_nodes/ast.ImportFrom` (same input node type).
    :raises UnexpandableImportStar: when `ReadPermissionError`,
        `UnparsableFile` or `ModuleNotFoundError` raised.
    """
    mpath = pathu.get_import_from_path(path, "*", node.module, node.level)

    importables: Set[str] = set()

    try:
        if mpath:
            content, _ = iou.safe_read(mpath, permissions=(os.R_OK,))
            tree = parse_ast(content, mpath)

            analyzer = ImportablesAnalyzer(mpath)
            analyzer.visit(tree)
            importables = analyzer.get_stats()
        else:
            name = ("." * node.level) + (node.module if node.module else "")
            raise ModuleNotFoundError(name=name)
    except (ReadPermissionError, UnparsableFile, ModuleNotFoundError) as err:
        msg = (
            err
            if not isinstance(err, ModuleNotFoundError)
            else f"{err.name!r} module not found or it's a C wrapped module!"
        )
        if hasattr(node, "location"):
            location = node.location  # type: ignore # pragma: nocover.
        else:
            location = _nodes.NodeLocation(
                (node.lineno, node.col_offset), 0  # type: ignore
            )
        raise UnexpandableImportStar(path, location, str(msg)) from err

    # Create `ast.alias` for each name.
    node.names.clear()
    for name in importables:
        node.names.append(ast.alias(name=name, asname=None))

    return node


def parse_ast(source_code: str, path: Path = Path(""), mode: str = "exec") -> ast.AST:
    """Parse the given `source_code` AST.

    :param source_code: python source code.
    :param path: `source_code` file path.
    :param mode: `ast.parse` mode.
    :returns: `ast.AST` (source code AST).
    :raises UnparsableFile: if the compiled source is invalid,
        or the source contains null bytes.
    """
    try:
        if PY38_PLUS:
            # Include type_comments when Python >=3.8.
            # For more information https://www.python.org/dev/peps/pep-0526/ .
            tree = ast.parse(source_code, mode=mode, type_comments=True)
        else:
            tree = ast.parse(source_code, mode=mode)
        return tree
    except (SyntaxError, IndentationError, ValueError) as err:
        raise UnparsableFile(path, err) from err



"""Pycln file IO utility."""
import os
import tokenize
from pathlib import Path
from typing import List, Tuple

from ._exceptions import ReadPermissionError, UnparsableFile, WritePermissionError


def safe_read(path: Path, permissions: tuple = (os.R_OK, os.W_OK)) -> Tuple[str, str]:
    """Read file content with encode detecting support.

    :param path: `.py` file path.
    :returns: decoded source code and file encoding.
    :raises ReadPermissionError: when `os.R_OK` in permissions
        and the source does not have read permission.
    :raises WritePermissionError: when `os.W_OK` in permissions
        and the source does not have write permission.
    :raises UnparsableFile: If both a BOM and a cookie are present, but disagree.
    """
    # Check these permissions before openinig the file.
    for permission in permissions:
        if not os.access(path, permission):
            if permission is os.R_OK:
                raise ReadPermissionError(13, "Permission denied [READ]", path)
            elif permission is os.W_OK:
                raise WritePermissionError(13, "Permission denied [WRITE]", path)
    try:
        with tokenize.open(path) as stream:
            source_code = stream.read()
            encoding = stream.encoding
        return source_code, encoding
    except SyntaxError as err:
        raise UnparsableFile(path, err) from err


def safe_write(path: Path, fixed_lines: List[str], encoding: str) -> None:
    """Write file content based on given `encoding`.

    :param path: `.py` file path.
    :param encoding: file encoding.
    :param fixed_lines: fixed source code lines.
    :raises WritePermissionError: when `os.W_OK` in permissions
        and the source does not have write permission.
    """
    if not os.access(path, os.W_OK):
        raise WritePermissionError(13, "Permission denied [WRITE]", path)
    with open(path, mode="w", encoding=encoding) as destination:
        destination.writelines(fixed_lines)



"""Pycln custom exceptions utility."""
from pathlib import Path
from typing import Union

from ._nodes import NodeLocation


class BaseOSError(Exception):

    """Custom OSError."""

    def __init__(self, errno: int, strerror: str, filepath: Path):
        message = f"{filepath} {strerror} [Errno {errno}]"
        super().__init__(message)


class ReadPermissionError(BaseOSError):

    """Raises when the file does not have read permission."""


class WritePermissionError(BaseOSError):

    """Raises when the file does not have write permission."""


class UnexpandableImportStar(Exception):

    """Raises when the import `*` statement unexpandable."""

    def __init__(self, path: Path, location: NodeLocation, msg: str):
        line, col = location.start.line, location.start.col
        message = f"{path}:{line}:{col} {self.__class__.__name__}: {msg}"
        super().__init__(message)


class UnparsableFile(Exception):

    """Raises when the compiled source code is invalid, or the source code
    contains null bytes."""

    def __init__(
        self,
        path: Path,
        err: Union[SyntaxError, IndentationError, ValueError],
    ):
        location = str(path)
        postfix = ""
        type_ = type(err)
        UnparsableFile._type_check(type_)

        if type_ in {SyntaxError, IndentationError}:
            lineno, col, text = err.lineno, err.offset, err.text  # type: ignore
            if lineno:
                location = f"{path}:{lineno}:{col}"
            if text:
                text = text.replace("\n", "").lstrip()
                postfix = f" {text!r}"

        elif type_ == ValueError:
            setattr(err, "msg", str(err))

        msg = err.msg  # type: ignore
        message = f"{location} {type_.__name__}: {msg}{postfix}"
        super().__init__(message)

    @staticmethod
    def _type_check(type_: type) -> None:
        """Validate the given `exception` type.

        :param type_: err type.
        :raises ValueError: if `type_` not in allowed types.
        """
        allowed_types = {SyntaxError, IndentationError, ValueError}
        if type_ not in allowed_types:
            raise ValueError(  # pragma: nocover
                f"UnparsableFile exception only takes {allowed_types}"
                + f" as err parameter but {type_!r} where given."
            )


class UnsupportedCase(Exception):

    """Raises when unsupported import case detected."""

    def __init__(self, path: Path, location: NodeLocation, msg: str):
        line, col = location.start.line, location.start.col
        cls_name = self.__class__.__name__
        message = f"{path}:{line}:{col} {cls_name}: {msg}"
        super().__init__(message)


def libcst_parser_syntax_error_message(path: Path, err) -> str:
    """Refactor `LibCST.ParserSyntaxError` message.

    :param path: where the exception has occurred.
    :param err: instance of `LibCST.ParserSyntaxError`.
    :returns: refactored message.
    """
    location = f"{path}:{err.raw_line}:{err.raw_column}"
    line = err._lines[0].replace("\n", "").lstrip()
    postfix = f" {line!r}"
    return f"{location} libcst.ParserSyntaxError: {err.message.rstrip('.')}:{postfix}"



"""Pycln code refactoring utility."""
import ast
import os
from functools import lru_cache
from importlib import import_module
from pathlib import Path
from typing import List, Optional, Set, Tuple, Union, cast

from . import iou, pathu, regexu, scan
from ._exceptions import (
    ReadPermissionError,
    UnexpandableImportStar,
    UnparsableFile,
    UnsupportedCase,
    WritePermissionError,
    libcst_parser_syntax_error_message,
)
from ._nodes import Import, ImportFrom, NodeLocation
from .config import Config
from .report import Report

# Constants.
NOPYCLN = "nopycln"
CHANGE_MARK = "\n_CHANGED_"
TRANSFORM = ".transform"
PYCLN_UTILS = "pycln.utils"


class LazyLibCSTLoader:

    """`transform.py` takes about '0.3s' to be loaded because of LibCST,
    therefore I've created this class to load it only if necessary.

    THIS CLASS DOES NOT INCLUDED ON THE TESTS SUITE. SO DON'T MODIFY IT
    FOR ANY REASON!
    """

    def __init__(self):
        self._module = None

    def __getattr__(self, name):
        if self._module is None:
            self._module = import_module(TRANSFORM, PYCLN_UTILS)
        return getattr(self._module, name)


transform = LazyLibCSTLoader()


class Refactor:

    """Refactor the given source.

    >>> refactor = Refactor(
    ...     configs,  # Should be created on the main function.
    ...     reporter,  # Should be created on the main function.
    ... )
    >>> file_path = "./source.py"
    >>> refactor.session(file_path)

    :param configs: `config.Config` instance.
    :param reporter: `report.Report` instance.
    """

    def __init__(self, configs: Config, reporter: Report):
        self.configs = configs
        self.reporter = reporter
        # Resetables.
        self._import_stats = scan.ImportStats(set(), set())
        self._source_stats = scan.SourceStats(set(), set(), set())
        self._path = Path("")

    def _reset(self) -> None:
        self._import_stats = scan.ImportStats(set(), set())
        self._source_stats = scan.SourceStats(set(), set(), set())
        self._path = Path("")

    @staticmethod
    def remove_useless_passes(source_lines: List[str]) -> List[str]:
        """Remove any useless `pass`.

        :param source_lines: source code lines.
        :returns: clean source code lines.
        """
        tree = ast.parse("".join(source_lines))
        for parent in ast.walk(tree):
            body = getattr(parent, "body", None)
            if body and hasattr(body, "__len__"):
                body_len = len(body)
            for child in ast.iter_child_nodes(parent):
                if isinstance(child, ast.Pass):
                    if body_len > 1:
                        body_len -= 1
                        source_lines[child.lineno - 1] = ""
        return "".join(source_lines).splitlines(True)

    def session(self, path: Path) -> None:
        """Refactoring session.

        Refactor the given `path` source code.

        :param path: `.py` file to refactor.
        """
        self._path = path
        try:
            # Safly read the file.
            permissions = [os.R_OK]
            if not self.configs.check and not self.configs.diff:
                permissions.append(os.W_OK)
            content, encoding = iou.safe_read(self._path, tuple(permissions))

            # Refactor and output the `content`.
            fixed_lines = self._code_session(content).splitlines(True)
            self._output(fixed_lines, content.splitlines(True), encoding)
        except (
            ReadPermissionError,
            WritePermissionError,
            UnparsableFile,
        ) as err:
            self.reporter.failure(str(err))
        finally:
            self._reset()

    def _code_session(self, source_code: str) -> str:
        """Refactor the given `source_code`.

        :param source_code: python source code.
        :returns: fixed source code.
        """
        # Skip any file that has `nopycln: file` comment.
        if regexu.skip_file(source_code):
            self.reporter.ignored_path(self._path, NOPYCLN)
            return source_code

        # Parse and analyze the `source_code` AST.
        tree = scan.parse_ast(source_code, self._path)
        original_lines = source_code.splitlines(True)
        stats = self._analyze(tree, original_lines)
        if not stats:
            return source_code
        self._source_stats, self._import_stats = stats

        # Refactor the `source_code`.
        return self._refactor(original_lines)

    def _output(
        self, fixed_lines: List[str], original_lines: List[str], encoding: str
    ) -> None:
        """Output the given `fixed_lines`.

        :param fixed_lines: the refactored source lines.
        :param original_lines: unmodified source lines.
        :param encoding: file encoding.
        """
        if fixed_lines == original_lines:
            self.reporter.unchanged_file(self._path)
        else:
            self.reporter.changed_file(self._path)
            if not self.configs.check:
                fixed_lines = Refactor.remove_useless_passes(fixed_lines)
                if self.configs.diff:
                    self.reporter.colored_unified_diff(
                        self._path, original_lines, fixed_lines
                    )
                else:
                    iou.safe_write(self._path, fixed_lines, encoding)

    def _analyze(
        self, tree: ast.AST, original_lines: List[str]
    ) -> Tuple[scan.SourceStats, scan.ImportStats]:
        """Analyze the given `tree`.

        :param tree: a parsed `ast.AST`.
        :param original_lines: code lines requiered for Python < 3.8.
        :returns: tuple of `ImportStats`, `SourceStats` and set of names to skip.
        """
        try:
            analyzer = scan.SourceAnalyzer(original_lines)
            analyzer.visit(tree)
            source_stats, import_stats = analyzer.get_stats()
            return source_stats, import_stats
        except Exception as err:
            self.reporter.failure(str(err), self._path)
            return None

    def _refactor(self, original_lines: List[str]) -> str:
        """Remove all unused imports from given `original_lines`.

        :param original_lines: unmodified lines.
        :reutrns: fixed source code.
        """
        fixed_lines = original_lines.copy()
        for type_ in self._import_stats:

            for node in type_:

                # Skip any import that has `# noqa` or `# nopycln: import` comment.
                lineno = node.location.start.line - 1
                if regexu.skip_import(fixed_lines[lineno]):
                    self.reporter.ignored_import(self._path, node)
                    continue

                # Expand any import '*' before checking.
                node, is_star = self._expand_import_star(node)
                if is_star is None:
                    continue

                # Get set of used names.
                used_names = self._get_used_names(node, is_star)

                # Depends on `--expand-stars, -x` option.
                if is_star:
                    if used_names:
                        if not self.configs.expand_stars:
                            continue
                        self.reporter.expanded_star(self._path, node)
                    else:
                        star_alias = ast.alias(name="*", asname=None)
                        self.reporter.removed_import(self._path, node, star_alias)

                # No alias has removed/added.
                if len(used_names) == len(node.names):
                    if not self.configs.expand_stars:
                        continue

                # Depends on `--check, -c` option.
                if self.configs.check:
                    fixed_lines.append(CHANGE_MARK)
                    continue

                # Default and `--diff, -d` option.
                fixed_lines = self._transform(
                    node.location, used_names, original_lines, fixed_lines
                )

        return "".join(fixed_lines)

    def _get_used_names(
        self, node: Union[Import, ImportFrom], is_star: bool
    ) -> Set[str]:
        """Get set of used names base on given `node` and `self._source_stats`.

        :param node: import node to names check.
        :parma is_star: is '*' import node.
        :returns: set of used names.
        """
        used_names: Set[str] = set()
        for alias in node.names:
            if self._should_remove(node, alias, is_star):
                if not is_star:
                    self.reporter.removed_import(self._path, node, alias)
                continue
            used_names.add(alias.name)
        return used_names

    def _transform(
        self,
        location: NodeLocation,
        used_names: Set[str],
        original_lines: List[str],
        updated_lines: List[str],
    ) -> List[str]:
        """Rebuild and replace the import without any unused part.

        :param location: `node.location`.
        :param used_names: set of all used names.
        :param original_lines: file original code lines.
        :param updated_lines: code lines to modify.
        :returns: modified source lines (fixed lines).
        """
        try:
            try:
                lineno = location.start.line - 1
                end_lineno = location.end.line
                import_stmnt = "".join(original_lines[lineno:end_lineno])
                rebuilt_import = transform.rebuild_import(
                    import_stmnt,
                    used_names,
                    self._path,
                    location,
                )
                updated_lines = Refactor._insert(
                    rebuilt_import, updated_lines, location
                )
            except UnsupportedCase as errin:
                self.reporter.failure(str(errin))
        except transform.cst.ParserSyntaxError as err:
            msg = libcst_parser_syntax_error_message(self._path, err)
            self.reporter.failure(msg)
        return updated_lines

    def _expand_import_star(
        self, node: ImportFrom
    ) -> Tuple[ImportFrom, Optional[bool]]:
        """Expand import star statement, `scan.expand_import_star` abstraction.

        :param node: `ImportFrom` that has a '*' as `alias.name`.
        :returns: expanded '*' import or the original node and True if it's star import.
        """
        try:
            is_star = False
            if node.names[0].name == "*":
                is_star = True
                node = cast(ImportFrom, scan.expand_import_star(node, self._path))
            return node, is_star
        except UnexpandableImportStar as err:
            self.reporter.failure(str(err))
            self.reporter.ignored_import(self._path, node, is_star=True)
            return node, None

    def _should_remove(
        self, node: Union[Import, ImportFrom], alias: ast.alias, is_star: bool
    ) -> bool:
        """Check if the alias should be removed or not.

        :param node: an `Import` or `ImportFrom`.
        :param alias: an `ast.alias` node.
        :param is_star: is it a '*' import.
        :returns: True if the alias should be removed else False.
        """
        real_name = node.module if isinstance(node, ImportFrom) else alias.name
        used_name = alias.asname if alias.asname else alias.name
        if (
            not self._has_used(used_name, is_star)
            and real_name not in pathu.IMPORTS_WITH_SIDE_EFFECTS
        ):
            if self.configs.all_ or real_name in pathu.get_standard_lib_names():
                return True
            if self._has_side_effects(alias.name, node) in (
                scan.HasSideEffects.NO,
                scan.HasSideEffects.NOT_MODULE,
            ):
                if (
                    not real_name
                    or self._has_side_effects(real_name, node) is scan.HasSideEffects.NO
                ):
                    return True
        return False

    def _has_used(self, name: str, is_star: bool) -> bool:
        """Check if the given import name has used.

        :param name: a name to check.
        :param is_star: is it a '*' import.
        :returns: True if the name has used else False.
        """
        name = name.split(".") if "." in name else name  # type: ignore
        if isinstance(name, str):
            if is_star and name in self._source_stats.names_to_skip:
                return False
            # Handle imports like (import os, from os import path).
            return name in self._source_stats.name_
        else:
            # Handle imports like (import os.path, from os import path.join).
            return self._has_used(name[0], is_star) and all(
                [name in self._source_stats.attr_ for name in name[1:]]
            )

    @lru_cache()
    def _has_side_effects(
        self, module: str, node: Union[Import, ImportFrom]
    ) -> scan.HasSideEffects:
        """Check if the given import file tree has side effects.

        :param module: `alias.name` to check.
        :param node: an `ast.Import` or `ast.ImportFrom` node.
        :returns: side effects status.
        """
        if isinstance(node, ImportFrom):
            module_source = pathu.get_import_from_path(  # pragma: nocover
                self._path, module, node.module, node.level
            )
        else:
            module_source = pathu.get_import_path(self._path, module)

        if not module_source:
            return scan.HasSideEffects.NOT_MODULE

        try:
            code, _ = iou.safe_read(module_source, permissions=(os.R_OK,))
            tree = scan.parse_ast(code, module_source)
        except (ReadPermissionError, UnparsableFile) as err:
            self.reporter.failure(str(err))
            return scan.HasSideEffects.NOT_KNOWN

        try:
            analyzer = scan.SideEffectsAnalyzer()
            analyzer.visit(tree)
            return analyzer.has_side_effects()
        except Exception as err:
            self.reporter.failure(str(err), self._path)
            return scan.HasSideEffects.NOT_KNOWN

    @staticmethod
    def _insert(
        rebuilt_import: List[str],
        updated_lines: List[str],
        location: NodeLocation,
    ) -> List[str]:
        """Insert (replace) rebuilt import statement into `updated_lines`.

        :param rebuilt_import: an import statement ot insert.
        :param updated_lines: a list of source lines to modify.
        :param location: unmodified node location.
        :returns: fixed list of lines.
        """
        # Shollow copy.
        fixed_lines = updated_lines.copy()

        # Determine old-new import delta.
        new_len = len(rebuilt_import)
        old_len = len(location)
        delta = old_len - new_len

        # Insert the rebuilt import statement.
        index = location.start.line - 1
        for i in range(new_len):
            if old_len == 1 and i != (new_len - 3):
                line = "".join(rebuilt_import[i:])
                fixed_lines[index] = line
                break
            else:
                fixed_lines[index] = rebuilt_import[i]
            index += 1
            old_len -= 1

        # Replace each removed line with `""`.
        if delta > 0:
            for i in range(delta):
                fixed_lines[index] = ""
                index += 1

        return fixed_lines



"""Pycln regex utility."""
import os
import re
import tokenize
from pathlib import Path
from typing import List, Pattern

import typer
from pathspec import PathSpec
from pathspec.patterns import GitWildMatchPattern

# Constants.
INCLUDE = "include"
EXCLUDE = "exclude"
GITIGNORE = ".gitignore"
SKIP_FILE_REGEX = r"# *(nopycln *: *file).*"
SKIP_IMPORT_REGEX = r"# *((noqa *:*)|(nopycln *: *import)).*"
INCLUDE_REGEX = r".*\.py$"
EXCLUDE_REGEX = (
    r"(\.eggs|\.git|\.hg|\.mypy_cache|__pycache__|\.nox|"
    + r"\.tox|\.venv|\.svn|buck-out|build|dist)/"
)


def safe_compile(pattern: str, type_: str) -> Pattern[str]:
    """Safely compile [--include, --exclude] options regex.

    :param pattern: an str regex to be complied.
    :param type_: 'include' OR 'exclude'.
    :returns: complied regex.
    """
    try:
        if isinstance(pattern, str):
            compiled: Pattern[str] = re.compile(pattern, re.IGNORECASE)
            return compiled
        return pattern  # type: ignore
    except re.error as err:
        typer.secho(
            f"Invalid regular expression for {type_} given: {pattern!r} ⛔",
            fg=typer.colors.RED,
            err=True,
        )
        raise typer.Exit(1) from err


def is_included(name: str, regex: Pattern[str]) -> bool:
    """Check if the file/directory name match include pattern.

    :param name: file/directory name to check.
    :param regex: include regex pattern.
    :returns: True if the name match else False.
    """
    return bool(regex.fullmatch(name))


def is_excluded(name: str, regex: Pattern[str]) -> bool:
    """Check if the file/directory name match exclude pattern.

    :param name: file/directory name to check.
    :param regex: exclude regex pattern.
    :returns: True if the name match else False.
    """
    return bool(regex.fullmatch(name))


def get_gitignore(root: Path, no_gitignore: bool) -> PathSpec:
    """Return a PathSpec matching gitignore content, if present.

    :param root: root path to search for `.gitignore`.
    :param no_gitignore: `config.no_gitignore` value.
    :returns: PathSpec matching gitignore content, if present.
    """
    lines: List[str] = []
    if not no_gitignore:
        path = os.path.join(root, GITIGNORE)
        if os.path.isfile(path):
            if os.access(path, os.R_OK):
                with tokenize.open(path) as ignore_file:
                    lines = ignore_file.readlines()
    return PathSpec.from_lines(GitWildMatchPattern, lines)


def skip_import(line: str) -> bool:
    """Check if the lines has `# noqa` or `# nopycln: import` to skip.

    :param line: a line to check.
    :returns: True if it matches else False.
    """
    return bool(re.search(SKIP_IMPORT_REGEX, line, re.IGNORECASE))


def skip_file(src_code: str) -> bool:
    """Check if the src_code code has `nopycln: file` comment to skip.

    :param src_code: string source code to check.
    :returns: True if it matches else False.
    """
    return bool(re.search(SKIP_FILE_REGEX, src_code, re.IGNORECASE))



"""Pycln import statements nodes utility."""
import ast
from dataclasses import dataclass
from typing import List, Optional, Tuple


@dataclass(frozen=True)
class NodePosition:

    """Node position class."""

    #: Line numbers are 1-indexed.
    line: int

    #: Column numbers are 0-indexed.
    col: Optional[int] = None

    def __hash__(self):
        if self.col:
            return hash(self.line + self.col)
        return hash(self.line)


@dataclass
class NodeLocation:

    """Node location class.

    :param start: tuple of start line and col_offset.
    :param end: tuple of end line.
    """

    def __init__(self, start: Tuple[int, int], end: int):
        self.start = NodePosition(*start)
        self.end = NodePosition(end)

    def __hash__(self):
        return hash(hash(self.start) + hash(self.end))

    def __len__(self):
        return (self.end.line - self.start.line) + 1


@dataclass
class BaseImport:

    """Custom `ast.AST` node."""

    #: Location contains:
    #:  - `ast.AST.lineno`.
    #:  - `ast.AST.col_offset`.
    #:  - `ast.AST.end_lineno`.
    #:
    #: `ast.AST.end_col_offset` not included.
    location: NodeLocation

    #: `ast.Import.names`.
    names: List[ast.alias]


@dataclass
class Import(BaseImport):

    """Custom `ast.Import` node."""

    def __hash__(self):
        return hash(self.location)


@dataclass
class ImportFrom(BaseImport):

    """Custom `ast.ImportFrom` node."""

    #: `ast.ImportFrom.module`.
    module: Optional[str]

    #: `ast.ImportFrom.level`.
    level: int

    @property
    def relative_name(self) -> str:
        """Node relative name."""
        dots = "." * self.level
        if self.module:
            return f"{dots}{self.module}"
        return dots

    def __hash__(self):
        return hash(self.location)



"""Pycln path finding utility."""
import os
import sys
from distutils import sysconfig
from functools import lru_cache
from pathlib import Path
from typing import Generator, List, Optional, Pattern, Set

from pathspec import PathSpec

from .. import ISWIN
from . import regexu
from .report import Report

# Constants.
EXCLUDE = "exclude"
INCLUDE = "include"
GITIGNORE = "gitignore"
PY_EXTENSION = ".py"
__INIT__ = "__init__.py"
LIB_DYNLOAD = "Lib" if ISWIN else "lib-dynload"
SITE_PACKAGES = "site-packages"
DIST_PACKAGES = "dist-packages"
LIB_PY_EXTENSIONS = ("so", "py", "pyc")
BIN_PY_EXTENSIONS = ("so", "egg-info", "nspkg.pth")
BIN_IMPORTS = {  # In case they are built into CPython.
    "io",
    "os",
    "sys",
    "grp",
    "pwd",
    "json",
    "math",
    "time",
    "parser",
    "string",
    "operator",
    "datetime",
    "multiprocessing",
}
IMPORTS_WITH_SIDE_EFFECTS = {"this", "antigravity", "rlcompleter"}


def yield_sources(
    path: Path,
    include: Pattern[str],
    exclude: Pattern[str],
    gitignore: PathSpec,
    reporter: Report,
) -> Generator:
    """Yields `.py` paths to handle. Walk throw path sub-directories/files
    recursively.

    :param path: A path to start searching from.
    :param include: regex pattern to be included.
    :param exclude: regex pattern to be excluded.
    :param gitignore: gitignore PathSpec object.
    :param reporter: a `report.Report` object.
    :returns: generator of `.py` files paths.
    """
    if path.is_file():
        if str(path).endswith(PY_EXTENSION):
            yield path
            return
        return

    dirs: List[str] = []
    files: List[str] = []

    is_included, is_excluded = regexu.is_included, regexu.is_excluded

    scandir = os.scandir(path)
    for entry in scandir:

        # Skip symlinks.
        if entry.is_symlink():
            continue

        name = entry.name if entry.is_file() else f"{entry.name}/"
        entry_path = Path(os.path.join(path, name))

        # Compute exclusions.
        if is_excluded(name, exclude):
            reporter.ignored_path(entry_path, EXCLUDE)
            continue

        # Compute `.gitignore`.
        if gitignore.match_file(name):
            reporter.ignored_path(entry_path, GITIGNORE)
            continue

        # Directories.
        if entry.is_dir():
            dirs.append(name)
            continue

        # Files.
        if is_included(name, include):
            files.append(name)
        else:
            reporter.ignored_path(entry_path, INCLUDE)

    for name in files:
        yield Path(os.path.join(path, name))

    for dirname in dirs:

        dir_path = Path(os.path.join(path, dirname))

        yield from yield_sources(dir_path, include, exclude, gitignore, reporter)


@lru_cache()
def get_standard_lib_paths() -> Set[Path]:
    """Get paths to Python standard library modules.

    :returns: set of paths to Python standard library modules.
    """
    paths: Set[Path] = set()

    for is_plat_specific in [True, False]:

        # Get lib modules paths.
        lib_path = sysconfig.get_python_lib(
            standard_lib=True, plat_specific=is_plat_specific
        )

        for path in os.listdir(lib_path):
            paths.add(Path(os.path.join(lib_path, path)))

        # Get lib dynload modules paths, if exists.
        lib_dynload_path = os.path.join(lib_path, LIB_DYNLOAD)

        if os.path.isdir(lib_dynload_path):

            for path in os.listdir(lib_dynload_path):
                paths.add(Path(os.path.join(lib_dynload_path, path)))

    return paths


@lru_cache()
def get_standard_lib_names() -> Set[str]:
    """Returns a set of Python standard library modules names.

    :returns: a set of Python standard library modules names.
    """
    names: Set[str] = set()
    paths: Set[Path] = get_standard_lib_paths()

    for path in paths:

        name = str(path.parts[-1])

        if name.startswith("_") or "-" in name:
            continue

        if "." in name and not name.endswith(LIB_PY_EXTENSIONS):
            continue

        names.add(name.split(".")[0])

    return (names - IMPORTS_WITH_SIDE_EFFECTS) | BIN_IMPORTS


@lru_cache()
def get_third_party_lib_paths() -> Set[Path]:
    """Get paths to third party library modules.

    :returns: set of paths to third party library modules.
    """
    paths: Set[Path] = set()

    packages_paths: Set[str] = {
        path
        for path in sys.path
        if path and Path(path).parts[-1] in [DIST_PACKAGES, SITE_PACKAGES]
    }

    for path in packages_paths:

        for name in os.listdir(path):
            if not name.startswith("_") and not name.endswith(BIN_PY_EXTENSIONS):
                paths.add(Path(os.path.join(path, name)))

    return paths


def get_local_import_path(path: Path, module: str) -> Optional[Path]:
    """Find the given local module file.py/__init_.py path.

    Written FOR `ast.Import`.

    :param path: where `module` has imported.
    :param module: module name.
    :returns: a full `module/__init__.py` path.
    """
    dirnames = Path(os.path.dirname(path)).parts
    names = module.split(".")

    # Test different levels.
    for i in [None] + list(range(-10, -0)):  # type: ignore

        # If it's a file.
        fpath = os.path.join(*dirnames[:i], *names[:-1], f"{names[-1]}{PY_EXTENSION}")
        if os.path.isfile(fpath):
            return Path(fpath)

        # If it's a module.
        mpath = os.path.join(*dirnames[:i], *names, __INIT__)
        if os.path.isfile(mpath):
            return Path(mpath)

    # Path not found.
    return None


def get_local_import_from_path(
    path: Path, module: str, package: str, level: int
) -> Optional[Path]:
    """Find the given local module file.py/__init_.py path.

    Written FOR `ast.ImportFrom`.

    :param path: where `module` has imported.
    :param module: module name.
    :param package: package name.
    :param level: `ast.ImportFrom.level`.
    :returns: a full `module/__init__.py` path.
    """
    dirname = Path(os.path.dirname(path))
    dirparts = dirname.parts[: (level * -1) + 1] if level > 1 else dirname.parts
    modules = module.split(".") if module != "*" and module else []
    packages = package.split(".") if package else []

    # Test different levels.
    for i in [None] + list(range(-10, -0)):  # type: ignore
        # If it's a file.
        if modules:
            fpath = os.path.join(
                *dirparts[:i],
                *packages,
                *modules[:-1],
                f"{modules[-1]}{PY_EXTENSION}",
            )
        else:
            # IMPORT "*" CASE.
            fpath = os.path.join(
                *dirparts[:i],
                *packages[:-1] if level > 0 else "",
                f"{packages[-1]}{PY_EXTENSION}",
            )
        if os.path.isfile(fpath):
            return Path(fpath)

        # If it's a module.
        if modules:
            mpath = os.path.join(
                *dirparts[:i],
                *packages,
                *modules,
                __INIT__,
            )
        else:
            # IMPORT "*" CASE.
            mpath = os.path.join(
                *dirparts[:i],
                *packages,
                __INIT__,
            )

        if os.path.isfile(mpath) and package.split(".")[0] in mpath:
            return Path(mpath)

    # Path not found.
    return None


def get_module_path(paths: Set[Path], module: str) -> Optional[Path]:
    """Get the `module` path from the given `paths`.

    :param paths: a list of paths to search.
    :param module: module name.
    :returns: `module` path if exist else None.
    """
    module = module.split(".")[0]
    for path in paths:
        name = str(path.parts[-1]).split(".")[0]
        if name == module:
            if str(path).endswith(PY_EXTENSION):
                return path
            else:
                return Path(os.path.join(path, __INIT__))
    # Path not found.
    return None


@lru_cache()
def get_import_path(path: Path, module: str) -> Optional[Path]:
    """Find the given module file.py/__init__.py path.

    Written for `ast.Import` nodes.

    :param path: where module has imported.
    :param module: module name.
    :returns: `module` file.py/__init_.py path, if found else None.
    """
    mpath = get_local_import_path(path, module)
    if mpath:
        return mpath

    elif module in get_standard_lib_names():
        return get_module_path(get_standard_lib_paths(), module)

    else:
        return get_module_path(get_third_party_lib_paths(), module)


@lru_cache()
def get_import_from_path(
    path: Path, module: str, package: str, level: int
) -> Optional[Path]:
    """Find the given module file.py/__init_.py path.

    Written for `ast.ImportFrom` nodes.

    :param path: where module has imported.
    :param module: module name.
    :param package: package name.
    :param level: `ast.ImportFrom.level`.
    :returns: `module` file.py/__init_.py path, if found else None.
    """
    mpath = get_local_import_from_path(path, module, package, level)
    if mpath:
        return mpath

    if module == "*":
        module = package

    if module in get_standard_lib_names():
        return get_module_path(get_standard_lib_paths(), module)

    elif package in get_standard_lib_names():
        return get_module_path(get_standard_lib_paths(), package)

    else:
        path = get_module_path(get_third_party_lib_paths(), module)
        if not path and package:
            path = get_module_path(get_third_party_lib_paths(), package)
        return path



import os
from pathlib import Path

import toml
import typer

ISWIN = os.name == "nt"
PYPROJECT_PATH = Path(__file__).parent.parent.joinpath("pyproject.toml")
pycln = toml.load(PYPROJECT_PATH)["tool"]["poetry"]

__name__ = pycln["name"]
__version__ = pycln["version"]
__doc__ = pycln["description"]


def version_callback(value: bool):
    """Show the version and exit with 0."""
    if value:
        typer.echo(f"{__name__}, version {__version__}")
        raise typer.Exit(0)



"""Pycln report utility."""
import ast
from dataclasses import dataclass
from difflib import unified_diff
from pathlib import Path
from typing import List, Optional, Union

import typer

from . import _nodes, config


@dataclass
class Report:

    """Provide a Pycln report counters.

    Can be rendered with `str(report)`.
    """

    #: Configured instance.
    configs: config.Config

    @staticmethod
    def get_location(path: Path, location: _nodes.NodeLocation) -> str:
        """Create full location from `path` and node location.

        :param path: file path.
        :param location: `_nodes.NodeLocation`.
        :returns: full location.
        """
        start = location.start
        line, col = str(start.line), str(start.col)
        return ":".join([str(path), line, col])

    @staticmethod
    def secho(
        message: str,
        *,  # Force kwargs.
        bold: bool,
        isedit: bool = False,
        issuccess: bool = False,
        iswarning: bool = False,
        iserror: bool = False,
    ) -> None:
        """Print a colored message.

        :param message: a string message.
        :param bold: is if a bold message.
        :param isedit: is it an edit message ~> stdout.
        :param issuccess: is it a success message  ~> stdout.
        :param iswarning: is it a warning message ~> stderr.
        :param iserror: is it an error message ~> stderr.
        """
        if isedit:
            color = typer.colors.BRIGHT_BLUE
        elif issuccess:
            color = typer.colors.BRIGHT_GREEN
        elif iswarning:
            color = typer.colors.BRIGHT_YELLOW
        elif iserror:
            color = typer.colors.BRIGHT_RED
        else:
            raise ValueError("Please specify one of the is* args.")
        # Print the colored message
        typer.echo(
            typer.style(" ", bg=color) + " " + typer.style(message, bold=bold),
            err=bool(iswarning or iserror),
        )

    @staticmethod
    def colored_unified_diff(
        path: Path,
        original_lines: List[str],
        fixed_lines: List[str],
    ) -> None:
        """Writeout colored and normalized diff.

        :param path: a file path.
        :param original_lines: original path code lines.
        :param fixed_lines: fixed path code lines.
        """
        diff_gen = unified_diff(
            original_lines,
            fixed_lines,
            fromfile=f"original/ {path}",
            tofile=f"fixed/ {path}",
            n=3,
            lineterm="\n",
        )
        diff_str = ""
        for line in diff_gen:
            if line.startswith("-"):
                line = typer.style(line, fg=typer.colors.RED)
            elif line.startswith("+"):
                line = typer.style(line, fg=typer.colors.GREEN)
            elif line.startswith("@"):
                line = typer.style(line, fg=typer.colors.CYAN)
            diff_str += line
        diff_str = diff_str.rstrip("\n ") + "\n"
        typer.echo(diff_str)

    @staticmethod
    def rebuild_report_import(
        node: Union[_nodes.Import, _nodes.ImportFrom],
        alias: ast.alias,
    ) -> str:
        """Rebuild import statement from AST for reporting purposes.

        :param node: import node.
        :param alias: `ast.alias` node.
        """
        str_alias = f"{alias.name} as {alias.asname}" if alias.asname else alias.name
        str_import = (
            f"from {node.relative_name} import"
            if isinstance(node, _nodes.ImportFrom)
            else "import"
        )
        return f"{str_import} {str_alias}"

    #: Total removed import statements counter.
    _removed_imports: int = 0

    def removed_import(
        self,
        path: Path,
        node: Union[_nodes.Import, _nodes.ImportFrom],
        removed_alias: ast.alias,
    ) -> None:
        """Increment `self._removed_imports`. Write a message to stdout.

        :param path: where the import has removed.
        :param node: removed import node.
        :param removed_alias: the removed `ast.alias` from the node.
        """
        if not any([self.configs.diff, self.configs.quiet, self.configs.silence]):
            location = Report.get_location(path, node.location)
            statement = Report.rebuild_report_import(node, removed_alias)
            removed = "whould be removed" if self.configs.check else "has removed"
            Report.secho(
                f"{location} {statement!r} {removed}! 🔮",
                bold=False,
                isedit=True,
            )

        self._removed_imports += 1
        self._file_removed_imports += 1 if self._file_removed_imports != -1 else 2

    #: Total expanded import statements counter.
    _expanded_stars: int = 0

    def expanded_star(self, path: Path, node: _nodes.ImportFrom) -> None:
        """Increment `self._expanded_stars`. Write a message to stdout.

        :param path: where the import has expanded.
        :param node: the expanded node.
        """
        if not any([self.configs.diff, self.configs.quiet, self.configs.silence]):
            location = Report.get_location(path, node.location)
            star_alias = ast.alias(name="*", asname=None)
            statement = Report.rebuild_report_import(node, star_alias)
            expanded = "whould be expanded" if self.configs.check else "has expanded"
            Report.secho(
                f"{location} {statement!r} {expanded}! 🔗",
                bold=False,
                isedit=True,
            )

        self._expanded_stars += 1
        self._file_expanded_stars += 1 if self._file_expanded_stars != -1 else 2

    #: Total changed files counter.
    _changed_files: int = 0

    #: These counters will be reseted for each file.
    _file_removed_imports: int = 0
    _file_expanded_stars: int = 0

    def _reset_file_counters(self) -> None:
        self._file_removed_imports = 0
        self._file_expanded_stars = 0

    def changed_file(self, path: Path) -> None:
        """Increment `self._changed_files`. Write a message to stdout.

        :param path: the changed file path.
        """
        if not any([self.configs.diff, self.configs.silence]):
            file_report: List[str] = []

            if self._file_removed_imports > 0:
                removed = "whould be removed" if self.configs.check else "has removed"
                s = "s" if self._file_removed_imports > 1 else ""
                file_report.append(f"{self._file_removed_imports} import{s} {removed}")

            if self._file_expanded_stars > 0:
                expanded = (
                    "whould be expanded" if self.configs.check else "has expanded"
                )
                s = "s" if self._file_expanded_stars > 1 else ""
                file_report.append(f"{self._file_expanded_stars} import{s} {expanded}")

            str_file_report = ", ".join(file_report)
            Report.secho(f"{path} {str_file_report}! 🚀", bold=True, isedit=True)

        self._changed_files += 1
        self._reset_file_counters()

    #: Total unchanged files counter.
    _unchanged_files: int = 0

    def unchanged_file(self, path: Path) -> None:
        """Increment `self._unchanged_files`. Write a message to stdout.

        :param path: the unchanged file path.
        """
        if self.configs.verbose and self._file_removed_imports != -1:
            Report.secho(f"{path} looks good! ✨", bold=False, issuccess=True)

        self._unchanged_files += 1
        self._reset_file_counters()

    #: Total ignored paths counter.
    _ignored_paths: int = 0

    def ignored_path(self, ignored_path: Path, type_: str) -> None:
        """Increment `self._ignored_paths`. Write a message to stderr.

        :param ignored_path: the ignored path.
        :param type_: ignore type (`exclude`, `include`, `gitignore` or `nopycln`).
        """
        if self.configs.verbose:
            if type_ == "exclude":
                type_ = "matches the --exclude regex"  # pragma: nocover.
            elif type_ == "gitignore":
                type_ = "matches the .gitignore patterns"  # pragma: nocover.
            elif type_ == "include":
                type_ = "does not match the --include regex"  # pragma: nocover.
            else:
                sharp = "#"  # To no skip this file.
                type_ = f"do to `{sharp} nopycln: file` comment"
            Report.secho(
                f"{ignored_path} has ignored: {type_}! ⚠️",
                bold=False,
                iswarning=True,
            )

        self._ignored_paths += 1

    #: Total ignored import statements counter.
    _ignored_imports: int = 0

    def ignored_import(
        self,
        path: Path,
        node: Union[_nodes.Import, _nodes.ImportFrom],
        is_star: bool = False,
    ) -> None:
        """Increment `self._ignored_imports`. Write a message to stderr.

        :param path: where the import has ignored.
        :param node: the ignored import node.
        :param is_star: set to true if it's a '*' import.
        """
        if self.configs.verbose:
            location = Report.get_location(path, node.location)
            statement = Report.rebuild_report_import(node, node.names[0]) + ("." * 3)
            reason = (
                "`# noqa` or `# nopycln: import`"
                if not is_star
                else "cannot expand the '*'"
            )
            Report.secho(
                f"{location} {statement!r} has ignored: {reason}! ⚠️",
                bold=False,
                iswarning=True,
            )
        self._ignored_imports += 1

    #: Total number of failures.
    _failures: int = 0

    def failure(self, msg: str, path: Optional[Path] = None) -> None:
        """Increment `self._failures`. Write a msg to stderr.

        :param msg: a failure msg.
        :param path: where the failure has appeared.
        """
        if not self.configs.silence:
            message = f"{path} {msg} ⛔" if path else f"{msg} ⛔"
            Report.secho(message, bold=False, iserror=True)
        if self._file_removed_imports == 0:
            self._file_removed_imports = -1
        self._failures += 1

    @property
    def exit_code(self) -> int:
        """Return an exit code.

        :returns: an exit code (0, 1, 250).
        """
        # According to http://tldp.org/LDP/abs/html/exitcodes.html
        # exit codes 1 - 2, 126 - 165, and 255 have special meanings,
        # and should therefore be avoided for user-specified exit parameters
        if self.configs.check:
            if self._failures:
                # Internal error (check).
                return 250
            if self._changed_files:
                # File(s) should be refactored.
                return 1
        else:
            if self._failures:
                # Internal error (not-check).
                return 1
        # Has modified or looks good.
        return 0

    @property
    def report_prefix(self) -> str:
        """Return the correct prefix.

        :returns: `"\n"` or `""`.
        """
        return (  # pragma: nocover.
            "\n"
            if any(
                [
                    self._changed_files,
                    all(
                        [
                            self.configs.verbose,
                            any([self._ignored_paths, self._ignored_imports]),
                        ]
                    ),
                    all(
                        [
                            any(
                                [
                                    self._failures,
                                    self._removed_imports,
                                    self._expanded_stars,
                                ]
                            ),
                            not self.configs.quiet,
                        ]
                    ),
                ]
            )
            else ""
        )

    def __str__(self) -> str:
        """Render a colored report of the current state.

        Use `typer.unstyle` to remove colors.

        :returns: a colored report of the current state.
        """
        if self.configs.silence:
            return ""

        if not any([self._changed_files, self._unchanged_files, self._failures]):
            typer.secho(
                ("\n" if self.configs.verbose and self._ignored_paths else "")
                + "No Python files are present to be cleaned. Nothing to do 😴",
                bold=True,
            )
            raise typer.Exit(1)

        if any([self.configs.check, self.configs.diff]):
            removed_imports = "would be removed"
            expanded_stars = "would be expanded"
            changed_files = "would be changed"
            unchanged_files = "would be left unchanged"
        else:
            removed_imports = "has removed"
            expanded_stars = "has expanded"
            changed_files = "has changed"
            unchanged_files = "left unchanged"
        failures = "has failed to be cleaned"

        report = []

        if self._removed_imports:
            s = "s" if self._removed_imports > 1 else ""
            report.append(
                typer.style(
                    f"{self._removed_imports} import{s} {removed_imports}",
                    bold=True,
                )
            )

        if self._expanded_stars:
            s = "s" if self._expanded_stars > 1 else ""
            report.append(
                typer.style(
                    f"{self._expanded_stars} import{s} {expanded_stars}",
                    bold=True,
                )
            )

        if self._changed_files:
            s = "s" if self._changed_files > 1 else ""
            report.append(
                typer.style(
                    f"{self._changed_files} file{s} {changed_files}",
                    bold=True,
                )
            )

        if self._unchanged_files:
            s = "s" if self._unchanged_files > 1 else ""
            report.append(
                typer.style(
                    f"{self._unchanged_files} file{s} {unchanged_files}",
                    bold=False,
                )
            )

        if self._failures:
            s = "s" if self._failures > 1 else ""
            report.append(
                typer.style(f"{self._failures} file{s} {failures}", bold=False)
            )

        if self.configs.verbose:
            ignored_imports = "has ignored"
            ignored_paths = "has ignored"

            if self._ignored_imports:
                s = "s" if self._ignored_imports > 1 else ""
                report.append(
                    typer.style(
                        f"{self._ignored_imports} import{s} {ignored_imports}",
                        bold=False,
                    )
                )

            if self._ignored_paths:
                s = "s" if self._ignored_paths > 1 else ""
                report.append(
                    typer.style(
                        f"{self._ignored_paths} path{s} {ignored_paths}",
                        bold=False,
                    )
                )

        if not self._failures:
            if self._removed_imports or self._expanded_stars:
                done_msg = "All done! 💪 😎"
            else:
                done_msg = "Looks good! ✨ 🍰 ✨"
        else:
            s = "are errors" if self._failures > 1 else "is an error"
            done_msg = f"Oh no, there {s}! 💔 ☹️"

        sdone_msg = typer.style(done_msg + "\n", bold=True)
        return self.report_prefix + sdone_msg + ", ".join(report) + ".\n"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pycln', 'pycln.utils']

package_data = \
{'': ['*']}

install_requires = \
['libcst>=0.3.10,<0.4.0',
 'pathspec>=0.8.0,<0.9.0',
 'pyyaml>=5.3.1,<6.0.0',
 'toml>=0.10.1,<0.11.0',
 'typer>=0.3.1,<0.4.0']

extras_require = \
{':python_version == "3.6"': ['dataclasses>=0.7,<0.8']}

entry_points = \
{'console_scripts': ['pycln = pycln.cli:app']}

setup_kwargs = {
    'name': 'pycln',
    'version': '0.0.1a3',
    'description': 'A formatter for finding and removing unused import statements.',
    'long_description': '<img src="https://raw.githubusercontent.com/hadialqattan/pycln/master/docs/_media/logo-background.png" width="100%" alt="Logo">\n\n<p align="center">\n    A formatter for finding and removing unused import statements.\n</p>\n\n<p align="center">\n    <a href="https://hadialqattan.github.io/pycln"><img src="https://img.shields.io/badge/For%20More%20Information%20See-Pycln%20Docs-B5FFB3.svg?style=flat-square" alt="Code style: prettier"></a>\n    <a href="https://github.com/hadialqattan/pycln/actions?query=workflow%3ACI"><img src="https://img.shields.io/github/workflow/status/hadialqattan/pycln/CI/master?label=CI&logo=github&style=flat-square" alt="CI"></a>\n    <a href="https://github.com/hadialqattan/pycln/actions?query=workflow%3ACD"><img src="https://img.shields.io/github/workflow/status/hadialqattan/pycln/CD?label=CD&logo=github&style=flat-square" alt="CD"></a>\n    <a href="https://www.codacy.com/manual/hadialqattan/pycln/dashboard?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=hadialqattan/pycln&amp;utm_campaign=Badge_Grade"><img src="https://img.shields.io/codacy/grade/e7c6c290c3c149e484634ac1905800d6/master?style=flat-square" alt="Codacy Badge"></a>\n    <a href="https://codecov.io/gh/hadialqattan/pycln"><img src="https://img.shields.io/codecov/c/gh/hadialqattan/pycln/master?token=VVYBDCZPHR&style=flat-square" alt="Codecov"></a>\n    <a href="https://codeclimate.com/github/hadialqattan/pycln/maintainability"><img src="https://img.shields.io/codeclimate/maintainability/hadialqattan/pycln?style=flat-square" alt="Maintainability"></a>\n</p>\n\n<p align="center">\n    <img src="https://img.shields.io/pypi/pyversions/pycln?style=flat-square" alt="PYPI - Python Version">\n    <a href="https://pypi.org/project/pycln/"><img src="https://img.shields.io/pypi/v/pycln?style=flat-square" alt="PYPI - Pycln Version"></a>\n    <a href="https://pypi.org/project/pycln/"><img src="https://img.shields.io/pypi/dm/pycln?color=dark-green&style=flat-square" alt="Downloads"></a>\n    <a href="https://hits.seeyoufarm.com"><img src="https://hits.seeyoufarm.com/api/count/incr/badge.svg?url=https%3A%2F%2Fgithub.com%2Fhadialqattan%2Fpycln&count_bg=%2344CC10&title_bg=%23555555&icon=&icon_color=%23E7E7E7&title=hits&edge_flat=true"/></a>\n    <a href="_blank"><img src="https://img.shields.io/tokei/lines/github.com/hadialqattan/pycln?style=flat-square" alt="Lines Of Code"></a>\n</p>\n\n<p align="center">\n    <a href="https://github.com/hadialqattan/pycln/fork"><img src="https://img.shields.io/github/forks/hadialqattan/pycln?style=flat-square" alt="Forks"></a>\n    <a href="https://github.com/hadialqattan/pycln/stargazers"><img src="https://img.shields.io/github/stars/hadialqattan/pycln?style=flat-square" alt="Stars"></a>\n    <a href="https://github.com/hadialqattan/pycln/issues"><img src="https://img.shields.io/github/issues/hadialqattan/pycln?style=flat-square" alt="Issues"></a>\n    <a href="https://github.com/hadialqattan/pycln/pulls"><img src="https://img.shields.io/github/issues-pr/hadialqattan/pycln?style=flat-square" alt="Pull Requests"></a>\n    <a href="https://github.com/hadialqattan/pycln/graphs/contributors"><img src="https://img.shields.io/github/contributors/hadialqattan/pycln?style=flat-square" alt="Contributors"></a>\n    <a href="https://github.com/hadialqattan/pycln/commits/master"><img src="https://img.shields.io/github/last-commit/hadialqattan/pycln.svg?style=flat-square" alt="Last Commit"></a>\n    <a href="https://github.com/hadialqattan/pycln/blob/master/LICENSE"><img src="https://img.shields.io/github/license/hadialqattan/pycln.svg?color=A31F34&style=flat-square" alt="License"></a>\n</p>\n\n<p align="center">\n    <a href="https://docutils.sourceforge.io/rst.html"><img src="https://img.shields.io/badge/docstrings-reStructuredText-gree.svg?style=flat-square" alt="Docstrings: reStructuredText"></a>\n    <a href="https://github.com/psf/black"><img src="https://img.shields.io/badge/code%20style-black-000000.svg?style=flat-square" alt="Code style: black"></a>\n    <a href="https://github.com/prettier/prettier"><img src="https://img.shields.io/badge/code%20style-prettier-ff69b4.svg?style=flat-square" alt="Code style: prettier"></a>\n</p>\n\n---\n\n<p align="center">\nFor more information see: https://hadialqattan.github.io/pycln\n</p>\n',
    'author': 'Hadi Alqattan',
    'author_email': 'alqattanhadizaki@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://hadialqattan.github.io/pycln',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<=3.9',
}


setup(**setup_kwargs)



"""Pycln configuration management utility."""
import configparser
import json
import tokenize
from dataclasses import dataclass
from pathlib import Path
from typing import List, Optional, Pattern, Union

import toml
import typer
import yaml

from . import regexu

# Constants.
CONFIG_SECTIONS = {
    ".cfg": "pycln",
    ".toml": "tool.pycln",
    ".json": "pycln",
    ".yaml": "pycln",
    ".yml": "pycln",
}


@dataclass
class Config:

    """Pycln configs dataclass."""

    def __post_init__(self):
        if self.config is not None:
            file_path = self.config
            self.config = None
            ParseConfigFile(file_path, self)
        else:
            self._check_path()
            self._check_regex()

    paths: List[Path]
    config: Optional[Path] = None
    include: Pattern[str] = regexu.INCLUDE_REGEX  # type: ignore
    exclude: Pattern[str] = regexu.EXCLUDE_REGEX  # type: ignore
    all_: bool = False
    check: bool = False
    diff: bool = False
    verbose: bool = False
    quiet: bool = False
    silence: bool = False
    expand_stars: bool = False
    no_gitignore: bool = False

    def _check_path(self) -> None:
        # Validate `self.paths`.
        if self.paths:
            for path in self.paths.copy():
                if not (path.is_dir() or path.is_file()):
                    self.paths.remove(path)

        if not self.paths:
            typer.secho(
                "No Path provided. Nothing to do 😴",
                bold=True,
                err=True,
            )
            raise typer.Exit(1)

    def _check_regex(self) -> None:
        # Validate `self.include/exclude`.
        self.include: Pattern[str] = regexu.safe_compile(
            str(self.include), regexu.INCLUDE
        )
        self.exclude: Pattern[str] = regexu.safe_compile(
            str(self.exclude), regexu.EXCLUDE
        )


class ParseConfigFile:

    """Conifg file parser.

    :param file_path: config file path.
    :param config: Config instance as base.
    """

    def __init__(self, file_path: Path, config: Config):
        self._path = file_path
        self._config = config
        self._section = CONFIG_SECTIONS.get(self._path.suffix, None)
        self.parse()
        self._config.__post_init__()

    def parse(self) -> None:
        """Get conifg from a `cfg`/`toml`/`json`/`yaml`/`yml` file."""
        if not self._path.is_file():
            typer.secho(
                f"Config file {str(self._path)!r} does not exist 😅",
                bold=True,
                err=True,
            )
            raise typer.Exit(1)
        if self._section is None:
            typer.secho(
                f"Config file {str(self._path)!r} is not supported 😅",
                bold=True,
                err=True,
            )
            typer.secho(f"Supported types: {CONFIG_SECTIONS.keys()}.", err=True)
            raise typer.Exit(1)
        getattr(self, f"_parse_{self._path.suffix.strip('.')}")()

    def _parse_cfg(self) -> None:
        # Parse `.cfg` file.
        parser = configparser.ConfigParser(allow_no_value=True)
        parser.read(self._path)
        cfg_data = parser._sections.get(self._section, {})  # type: ignore

        def cast_bool(v: str) -> Union[str, bool]:
            if v.lower() == "true":
                return True
            elif v.lower() == "false":
                return False
            return v

        configs = {k: cast_bool(v) for k, v in cfg_data.items()}
        self._config_loader(configs)

    def _parse_toml(self) -> None:
        # Parse `.toml` file.
        parsed_toml = toml.load(self._path)
        tool, pycln = self._section.split(".")
        configs = parsed_toml.get(tool, {}).get(pycln, {})
        self._config_loader(configs)

    def _parse_json(self) -> None:
        # Parse `.json` file.
        with tokenize.open(self._path) as stream:
            parsed_json = json.load(stream)
        configs = parsed_json.get(self._section, {})
        self._config_loader(configs)

    def _parse_yaml(self) -> None:
        # Parse `.yaml` file.
        with tokenize.open(self._path) as stream:
            parsed_yaml = yaml.load(stream, Loader=yaml.SafeLoader)
        configs = parsed_yaml.get(self._section, {})
        self._config_loader(configs)

    def _parse_yml(self) -> None:
        # Support `.yml` file.
        return self._parse_yaml()

    def _config_loader(self, config_dict: dict) -> None:
        # k, v: config loader.
        if config_dict:
            for k, v in config_dict.items():
                # Python preserved name.
                # `all` ~> `all_`.
                k = "all_" if k == "all" else k
                if hasattr(Config, k):
                    setattr(self._config, k, v)
